# from nltk.stem.wordnet import WordNetLemmatizer
# from nltk.corpus import twitter_samples, stopwords
# from nltk.tag import pos_tag
# from nltk.tokenize import word_tokenize
# from nltk import FreqDist, classify, NaiveBayesClassifier
# import pickle
# import time
# import os
# import re, string, random
#
# def remove_noise(tweet_tokens, stop_words = ()):
#
#     cleaned_tokens = []
#
#     for token, tag in pos_tag(tweet_tokens):
#         token = re.sub('http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+#]|[!*\(\),]|'\
#                        '(?:%[0-9a-fA-F][0-9a-fA-F]))+','', token)
#         token = re.sub("(@[A-Za-z0-9_]+)","", token)
#
#         if tag.startswith("NN"):
#             pos = 'n'
#         elif tag.startswith('VB'):
#             pos = 'v'
#         else:
#             pos = 'a'
#
#         lemmatizer = WordNetLemmatizer()
#         token = lemmatizer.lemmatize(token, pos)
#
#         if len(token) > 0 and token not in string.punctuation and token.lower() not in stop_words:
#             cleaned_tokens.append(token.lower())
#     return cleaned_tokens
#
# def predict(text):
#     dirname = os.path.dirname(__file__)
#     filename = 'sentiment_analysis_classifier.sav'
#     loaded_model = pickle.load(open(dirname + '/sentiment_analysis_classifier.sav', 'rb'))
#
#     custom_tokens = remove_noise(word_tokenize(text))
#
#     result = loaded_model.classify(dict([token, True] for token in custom_tokens))
#     return result


import unittest
from upswingutil.ml import cancellation_risk_score
import upswingutil as ul
from upswingutil.pms import PMS

ul.ENCRYPTION_SECRET = "S1335HwpKYqEk9CM0I2hFX3oXa5T2oU86OXgMSW4s6U="
ul.MONGO_URI = "mongodb://AdminUpSwingGlobal:Upswing098812Admin0165r@dev.db.upswing.global:27017/?authSource=admin&readPreference=primary&appname=Agent%20Oracle%20Dev&ssl=false"
ul.G_CLOUD_PROJECT = "aura-staging-31cae"
ul.FIREBASE = "/Users/harsh/upswing/github/api-oracle/SECRET/aura-staging-31cae-firebase-adminsdk-dyolr-7c135838e9.json"
ul.LOG_LEVEL_VALUE = 'DEBUG'


class TestCancellationRiskScore(unittest.TestCase):

    def test_creating_model_for_rms(self):
        """ Train and store model for RMS """
        orgId = '11249'
        crf = cancellation_risk_score(PMS.RMS, orgId)
        crf.preprocess()
        crf.train()



    def test_predict_from_model_rms(self):
        orgId = '11249'
        crf = cancellation_risk_score(PMS.RMS, orgId)
        result = crf.predict(9, 2, 1, 'unknown', 0, 104, 0, 11263, 0, 0, 'VIP', 'Duo', 52.0, 5.0, 78.0, 7, 11, 4.0)

        print(result)


if __name__ == '__main__':
    unittest.main()



# import pickle
# from util.db import Firestore, Mongodb
# from util.error_handling import log_error
# import pandas as pd
# from sklearn.cluster import KMeans
# from sklearn.preprocessing import StandardScaler
#
# __file_path__ = './ml/trained_models/booking_categorization'
#
#
# def train_k_means(clientId, df):
#     scaler = StandardScaler()
#     data_scaled = scaler.fit_transform(df)
#
#     SSE = []
#     for cluster in range(1, 20):
#         kmeans = KMeans(n_jobs=-1, n_clusters=cluster, init='k-means++')
#         kmeans.fit(data_scaled)
#         SSE.append(kmeans.inertia_)
#
#     kmeans = KMeans(n_jobs=-1, n_clusters=4, init='k-means++')
#     kmeansfit = kmeans.fit(data_scaled)
#
#     pickle.dump(kmeansfit, open(f'{__file_path__}_{clientId}.pkl', 'wb'))
#
#
# def transform_data(raw_data):
#     # raw_data1=raw_data[raw_data['status']=='Departed']
#     raw_data["departureDate"] = pd.to_datetime(raw_data["departureDate"])
#     raw_data["departureDate"]
#     raw_data["departureDate"].dt.date
#     raw_data["arrivalDate"] = pd.to_datetime(raw_data["arrivalDate"])
#     raw_data["arrivalDate"]
#     raw_data["arrivalDate"].dt.date
#     stay_d = (raw_data['departureDate'] - raw_data['arrivalDate']).dt.days
#     raw_data['stay_d'] = [int(i) for i in stay_d]
#     raw_data['total_rate'] = [float(i) for i in raw_data['totalRate']]
#     df = raw_data[
#         ['booking_type', 'children', 'duration_type', 'total_rate', 'guestId', 'stay_d', 'propertyId', 'categoryId']]
#
#     df = df[df['stay_d'] > 0]
#     df['average_perday'] = df['total_rate'] / df['stay_d']
#     df['guestId'] = [int(i) for i in df['guestId']]
#
#     df = df[df['stay_d'] < 35]
#     df = df[df['total_rate'] < 40000]
#
#     df_dummies = pd.get_dummies(df['booking_type'])
#     df_new = pd.concat([df, df_dummies], axis=1)
#     del df_new['booking_type']
#     df_new.head()
#
#     df = pd.get_dummies(df_new['categoryId'])
#     del df[df.columns[-1]]
#     df = pd.concat([df_new, df], axis=1)
#     del df['categoryId']
#     df.head()
#
#     df = df[["stay_d", "total_rate", 'Urgent Booking']]
#
#     return df
#
#
# def train_booking_categorization_model_by_org(app, clientId):
#     db = Firestore()
#     path = f'{db.org_collection}/{clientId}/{db.sub_collection_properties}'
#     props = db.get_collection(path).stream()
#     for prop in props:
#         train_booking_categorization_model(app, clientId, prop.id)
#
#
# def train_booking_categorization_model(app, dbName, clientId):
#     with app.app_context():
#         db = None
#         result = None
#         try:
#             app.logger.info(
#                 f'Starting training booking categorization model for property {clientId} and client {dbName}')
#             db = Mongodb(dbName)
#             __data_pipeline__ = [
#                 {
#                     '$match': {
#                         "clientId": int(clientId),
#                         'status': 'Departed'
#                     }
#                 },
#                 {
#                     '$project': {
#                         'status': 1, 'departureDate': 1, 'arrivalDate': 1,
#                         'totalRate': '$financial_info_actual.totalRate',
#                         'booking_type': 1, 'children': 1, 'duration_type': 1, 'guestId': 1, 'categoryId': 1,
#                         'propertyId': 1
#                     }
#                 }
#             ]
#             raw_data = pd.DataFrame(list(db.get_collection(db.reservation_collection).aggregate(__data_pipeline__)))
#             if not raw_data.empty:
#                 df = transform_data(raw_data)
#                 train_k_means(clientId, df)
#                 app.logger.info(f"New Model created, File saved: {__file_path__}_{clientId}.pkl")
#                 result = 'success'
#                 app.logger.info(f'Completed training booking categorization model for {clientId}')
#             else:
#                 app.logger.info(f'No sufficient data to train booking categorization model for {clientId}')
#
#         except Exception as e:
#             app.logger.error(e)
#             result = 'failed'
#             log_error(e, f'Failed to train booking categorization model for {clientId}')
#         finally:
#             if db:
#                 db.close_connection()
#             return result
#
#
# class BookingType:
#
#     @classmethod
#     def predict(cls, clientId, stay, spend, booking_type):
#         result = None
#         try:
#             booking_type = 1 if booking_type == 'Urgent Booking' else 0
#             model = pickle.load(open(f'{__file_path__}_{clientId}.pkl', 'rb'))
#             prediction = model.predict([[stay, spend, booking_type]])
#             result = prediction[0] if len(prediction) > 0 else None
#         except Exception as e:
#             log_error(e, f'Failed predicting booking type for {clientId}')
#         finally:
#             return result
#
#
# if __name__ == '__main__':
#     val = BookingType.predict(11290, '27', '30000', 'Pre planned')
#     print(val)



import pandas as pd
import numpy as np
from upswingutil.db import Mongodb
from loguru import logger
import tensorflow as tf
# from sklearn.model_selection import train_test_split
# from sklearn.utils import resam
from upswingutil.resource import get_model_from_cloud_storage, upload_model_to_cloud_storage


class CancellationRiskScore:
    __file_name__ = 'cancellation_risk_score_{}.pkl'

    def __init__(self, orgId: str, hotelId: str):
        self.orgId = orgId
        self.hotelId = hotelId

    def get_data(self) -> list:
        """
        retrieve data from mongodb for the model to train
        """
        __pipeline__ = [
            {
                '$match': {
                    'status': {
                        '$nin': [
                            'Quote', 'Unconfirmed', 'OwnerOccupied', 'Maintenance'
                        ]
                    }
                }
            },
            {
                '$addFields': {
                    'cancellation_lead_in': {
                        '$floor': {
                            '$divide': [
                                {
                                    '$subtract': [
                                        {
                                            '$toDate': '$arrivalDate'
                                        }, {
                                            '$toDate': '$cancelledDate'
                                        }
                                    ]
                                }, 60 * 60 * 24 * 1000
                            ]
                        }
                    }
                }
            },
            {
                '$addFields': {
                    'durationOfStay': {
                        '$floor': {
                            '$divide': [
                                {
                                    '$subtract': [
                                        {
                                            '$toDate': '$departureDate'
                                        }, {
                                            '$toDate': '$arrivalDate'
                                        }
                                    ]
                                }, 60 * 60 * 24 * 1000
                            ]
                        }
                    }
                }
            },
            {
                '$addFields': {
                    'Leadin': {
                        '$floor': {
                            '$divide': [
                                {
                                    '$subtract': [
                                        {
                                            '$toDate': '$arrivalDate'
                                        }, {
                                            '$toDate': '$createdDate'
                                        }
                                    ]
                                }, 60 * 60 * 24 * 1000
                            ]
                        }
                    }
                }
            },
            {
                '$project': {
                    '_id': 1,
                    'cancellation_lead_in': 1,
                    'durationOfStay': 1,
                    'Leadin': 1,
                    'geustCount': 1,
                    'clientId': 1,
                    'Class': 1,
                    'adults': 1,
                    'bookingSourceId': 1,
                    'booking_level': 1,
                    'categoryId': 1,
                    'children': 1,
                    "cancellationPolicyId": 1,
                    'cancellation_time_indays': '7',
                    'guest_type': 1,
                    'paymentModeId': 1,
                    'status': 1,
                    'travelAgentId': 1,
                    'ArrivalMonth': {
                        '$month': {
                            '$toDate': '$arrivalDate'

                        }
                    },
                    'dayoftheweek': {
                        '$dayOfWeek': {
                            '$toDate': '$arrivalDate'  # (according to dataset-- sunday is1 , sat is 7)
                        }
                    }

                }

            }
        ]
        mongo = Mongodb(self.orgId)
        result = mongo.execute_pipeline(mongo.RESERVATION_COLLECTION, __pipeline__)
        mongo.close_connection()
        return result

    def preprocess(self):
        """
        preprocess data extracted from mongodb and return data for training and testing of model
        """
        df = pd.DataFrame(self.get_data())
        logger.info(f'DataFrame Created from the Data Pipeline !!')

        df.dropna(axis=0, inplace=True)
        df['status'] = df['status'].apply(lambda x: 'Not Cancelled' if x in ['Departed', 'Arrived'] else 'Cancelled')

        # splitting the data----
        x = df.loc[:,
            ['_id', 'adults', 'bookingSourceId', 'booking_level', 'cancellationPolicyId', 'categoryId',
             'children', 'clientId', 'paymentModeId', 'travelAgentId', 'Class', 'guest_type',
             'cancellation_lead_in', 'durationOfStay', 'Leadin', 'cancellation_time_indays', 'ArrivalMonth',
             'dayoftheweek']]
        y = df.loc[:, ['status']]

        # Encoding----
        from sklearn.preprocessing import LabelEncoder
        le = LabelEncoder()
        y = le.fit_transform(y)

        # Encoding----
        from sklearn.compose import ColumnTransformer
        from sklearn.preprocessing import OneHotEncoder
        ct = ColumnTransformer(transformers=[('encoder', OneHotEncoder(), [3, 10, 11])],
                               remainder="passthrough")
        x = ct.fit_transform(x)

        from sklearn.model_selection import train_test_split
        x_train, x_test, y_train, y_test = train_test_split(x, y, test_size=0.2, random_state=0)

        return x_train, x_test, y_train, y_test

    def train(self):
        file_name = CancellationRiskScore.__file_name__.format(self.hotelId)
        """ Create model using preprocessed data and store the model in google cloud storage """
        x_train, x_test, y_train, y_test = self.preprocess()
        x_train = np.asarray(x_train).astype(
            int)  # as this error was coming (valueerror-failed-to-convert-a-numpy-array-to-a-     tensor-unsupported-object-type)
        y_train = np.asarray(y_train).astype(int)
        x_test = np.asarray(x_test).astype(int)

        # Initializing the ANN
        ann = tf.keras.models.Sequential()

        # Adding the input layer and the first hidden layer
        ann.add(tf.keras.layers.Dense(units=6, activation='relu'))

        # Adding the second hidden layer
        ann.add(tf.keras.layers.Dense(units=6, activation='relu'))

        # Adding the output layer
        ann.add(tf.keras.layers.Dense(units=1, activation='sigmoid'))

        # Part 3 - Training the ANN

        # Compiling the ANN
        ann.compile(optimizer='adam', loss='binary_crossentropy', metrics=['accuracy'])

        # Training the ANN on the Training set
        model = ann.fit(x_train, y_train, batch_size=32, epochs=100)
        upload_model_to_cloud_storage(model, self.orgId,
                                      file_name)  # storing the generated model as a pickle file in cloud storage
        logger.info(f"New Model created, File saved: {file_name}")
        logger.info(f"Completed Cancellation Risk Score model for {self.hotelId}")

    def predict(self, _id, adults, bookingSourceId, booking_level, cancellationPolicyId, categoryId,
                children, clientId, paymentModeId, travelAgentId, Class, guest_type,
                cancellation_lead_in, durationOfStay, Leadin, cancellation_time_indays, ArrivalMonth,
                dayoftheweek):
        """ predict result for the new data from the model created while training and stored in cloud storage """
        file_name = CancellationRiskScore.__file_name__.format(self.hotelId)
        model = get_model_from_cloud_storage(self.orgId, file_name)
        result = model.predict(_id, adults, bookingSourceId, booking_level, cancellationPolicyId, categoryId,
                               children, clientId, paymentModeId, travelAgentId, Class, guest_type,
                               cancellation_lead_in, durationOfStay, Leadin, cancellation_time_indays, ArrivalMonth,
                               dayoftheweek)

        return {
        'resId': _id,
        'score': round(result * 100, 2)
        }



from .reservation_count_forecasting import ReservationCountForecast
from .revenue_forecasting_v2 import RevenueForecasting
from .cancellation_risk_score import CancellationRiskScore
# from .classification_cancel_noShow_others import Classification_Cancellation_NoShow_Others
